module eigenstate_utils

IMPLICIT NONE

  DOUBLE PRECISION, PARAMETER :: dpi = datan(1.d0)*4.d0
  DOUBLE PRECISION, Parameter :: pi=acos(dble(-1.0))
  DOUBLE PRECISION, PARAMETER :: egam=0.57721566490153286060651209008240243104215933593992
  COMPLEX, Parameter          :: Ii=(0.0d0,1.0d0) ! csqrt(-1)
  double precision, parameter :: kbolzev = 8.6173324d-5 ! (*  eV K-1 *)
  double precision, parameter :: hartev = 27.21138386 
  double precision, parameter :: evj =  1.60218d-19! electronvolt to joule
  double precision, parameter :: hplanck = 6.6260755d-34 ! SI
  double precision, parameter :: clight =  2.997924d8   ! SI
  double precision, parameter :: bohrm = 5.29177249d-11; !conversion from bohr to meter
  double precision, parameter :: bohrangst = bohrm*1d10 !conversion from bohr to angst
  double precision, parameter :: umelec = 1822.888484514 !conversion from u to units of electron mass


contains


subroutine rombe ( func, a, b, aserr, relerr, error, result, ind )
!
!! CADRE estimates the integral of F(X) from A to B.
!  CADRE is the Cautious Adaptive Romberg Extrapolator.
!
!  Parameters:
!    Input, double precision, external FUNC, the name of the function to be integrated.
!    The user must declare the name an external parameter in the calling
!    program, write a function routine of the form FUNCTION FUNC(X) which
!    evaluates the function at X, and pass the name of the function
!    in FUNC.
  implicit none
  double precision, external :: func  
  double precision, intent(in) :: a ! lower limit of integration
  double precision, intent(in) :: b ! upper limit of integration
  double precision, intent(in) :: aserr ! absolute error tolerance
  double precision, intent(in) ::  relerr ! relative error tolerance
  double precision, intent(out)::  error ! estimate of the absolute error.
  double precision, intent(out)::  result ! approximate value of the integral.
  integer, intent(out) :: ind ! reliability indicator.
!    If IND <= 2, RESULT is very reliable.  Higher values of
!    IND indicate less reliable values of RESULT.
  integer, parameter :: mxstge = 30
  integer, parameter :: mxtbl = 10
  integer, parameter :: mxts = 2049
  integer iend, ji, iii,  istage, istep, istep2, it, l, lm1, n, n2, nnleft, i, ibeg
  integer ibegs(mxstge)
!
  double precision, parameter :: aitlow = 1.1d+00
  double precision, parameter :: aittol = 0.1d+00
  double precision, parameter :: h2tol = 0.15d+00
  double precision, parameter :: tljump = 0.01d+00
  
  double precision astep, beg, bma, curest, diff, end, ergoal, erra, errer,errr,fbeg, fbeg2, fend, fextm1, fextrp, fnsize,fn
  double precision h2next, h2tfex,hovn
  double precision rn(4), rnderr, sing, singnx, slope, stage, step, stepmn,prever,tabs, tabtlm,vint, sum1, sumabs
  double precision begin(mxstge), dif(mxtbl),finis(mxstge),est(mxstge),t(mxtbl,mxtbl), ts(2049)
  double precision r(mxtbl), ait(mxtbl)

  logical h2conv, aitken, reglar,right
  logical reglsv(mxstge)
!
  if ( a == b ) then
    result = 0.0d+00
    return
  end if
 
  begin(1:mxstge) = 0.0d+00
  est(1:mxstge) = 0.0d+00
  finis(1:mxstge) = 0.0d+00
  ibegs(1:mxstge) = 0
  reglsv(1:mxstge) = .false.
 
  vint = 0.0d+00
 
  rn(1:4) = (/ 0.7142005d+00, 0.3466282d+00, 0.8437510d+00, 0.1263305d+00 /)
 
  rnderr = epsilon ( rnderr )
  result = 0.0d+00
  error = 0.0d+00
  ind = 1
  bma = dabs ( b - a )
  errr = min ( 0.1d+00, dmax1( abs ( relerr ), 10.0d+00*rnderr) )
  erra = dabs ( aserr )
  stepmn = dmax1( bma / 2**mxstge, dmax1( bma, dabs ( a ), dabs ( b ) ) * rnderr )
  stage = 0.5
  istage = 1
  curest = 0.0d+00
  fnsize = 0.0d+00
  prever = 0.0d+00
  reglar = .false.
  beg = a
  fbeg = func(beg) / 2.0d+00
  ts(1) = fbeg
  ibeg = 1
  end = b
  fend = func(end) / 2.0d+00
  ts(2) = fend
  iend = 2
 
10 continue
 
  right = .false.
 
20 continue

  step = end - beg
  astep = abs ( step )
 
  if ( astep < stepmn ) then
    ind = 5
    result = dble(curest + vint)
    return
  end if
 
  t(1,1) = fbeg+fend
  tabs = abs ( fbeg ) + abs ( fend )
  l = 1
  n = 1
  h2conv = .false.
  aitken = .false.
  go to 40
 
30 continue
 
40 continue
 
  lm1 = l
  l = l+1
  n2 = n*2
  fn = n2
  istep = (iend-ibeg)/n

  if ( istep > 1 ) then
    go to 60
  end if

  ji = iend
  iend = iend + n

  if ( iend > mxts ) then
    go to 440
  end if

  hovn = step / fn
 
  iii = iend
  do i = 1, n2, 2
    ts(iii) = ts(ji)
    ts(iii-1) = func(end-i*hovn)
    iii = iii-2
    ji = ji-1
  end do
 
  istep = 2
 
60 continue
 
  istep2 = ibeg+istep/2
 
  sum1 = 0.0d+00
  sumabs = 0.0d+00
  do i = istep2, iend, istep
    sum1 = sum1 + ts(i)
    sumabs = sumabs + abs ( ts(i) )
  end do
 
  t(l,1) = t(l-1,1) / 2.0d+00 + sum1 / fn
  tabs = tabs / 2.0d+00 + sumabs / fn
 
  n = n2
  it = 1
  vint = step * t(l,1)
  tabtlm = tabs * rnderr
  fnsize = dmax1( fnsize, abs ( t(l,1) ) )
  ergoal = dmax1( astep * rnderr * fnsize, &
    stage * dmax1( erra , errr * abs ( curest+vint ) ) )
  fextrp = 1.0d+00
  do i = 1, lm1
    fextrp = fextrp * 4.0d+00
    t(i,l) = t(l,i) - t(l-1,i)
    t(l,i+1) = t(l,i) + t(i,l) / ( fextrp - 1.0 )
  end do
 
  errer = astep * abs ( t(1,l) )
  if ( l > 2 ) go to 90
  if ( abs ( t(1,2) ) <= tabtlm ) go to 290
  go to 40
 
90 continue
 
  do i = 2, lm1

    if ( abs ( t(i-1,l) ) > tabtlm ) then
      diff = t(i-1,lm1) / t(i-1,l)
    else
      diff = 0.0d+00
    end if

    t(i-1,lm1) = diff

  end do
 
  if ( abs ( 4.0 - t(1,lm1) ) <= h2tol ) go to 130
  if ( t(1,lm1) == 0.0 ) go to 120
  if ( abs ( 2.0 - abs ( t(1,lm1) ) ) < tljump ) go to 280
  if (l==3) go to 30
  h2conv = .false.
  if ( abs ( ( t(1,lm1) - t(1,l-2) ) / t(1,lm1) ) <= aittol ) go to 160
 
  if ( .not. reglar .and. l == 4 ) go to 30
 
120 continue
 
  if ( errer <= ergoal ) go to 310
  go to 380

130 continue

  if ( .not. h2conv ) then
    aitken = .false.
    h2conv = .true.
  end if

140 continue

  fextrp = 4.0d+00

150 continue

  it = it+1
  vint = step * t(l,it)
  errer = abs ( step / (fextrp-1.0) * t(it-1,l))
  if ( errer <= ergoal ) go to 340
  if ( it == lm1 ) go to 270
  if ( t(it,lm1) == 0.0 ) go to 150
  if ( t(it,lm1) <= fextrp ) go to 270

  if ( abs ( t(it,lm1) / 4.0 - fextrp ) / fextrp < aittol ) then
    fextrp = fextrp*4.0d+00
  end if

  go to 150
 
160 continue

  if ( t(1,lm1) < aitlow ) then
    go to 380
  end if
 
  if ( .not. aitken ) then
    h2conv = .false.
    aitken = .true.
  end if
 
170 continue

  fextrp = t(l-2,lm1)
  if ( fextrp > 4.5 ) go to 140
  if ( fextrp < aitlow ) go to 380

  if ( abs ( fextrp - t(l-3,lm1) ) / t(1,lm1) > h2tol ) then
    go to 380
  end if

  sing = fextrp
  fextm1 = fextrp - 1.0d+00

  ait(1) = 0.0d+00
  do i = 2, l
    ait(i) = t(i,1) + (t(i,1)-t(i-1,1)) / fextm1
    r(i) = t(1,i-1)
    dif(i) = ait(i) - ait(i-1)
  end do

  it = 2

190 continue

  vint = step*ait(l)

200 continue

  errer = errer / fextm1
 
  if ( errer <= ergoal ) then
    ind = max0 ( ind, 2 )
    go to 340
  end if
 
210 continue

  it = it+1
  if ( it == lm1 ) go to 270

  if ( it <= 3 ) then
    h2next = 4.0d+00
    singnx = 2.0d+00 * sing
  end if

  if ( h2next < singnx ) go to 230
  fextrp = singnx
  singnx = 2.0d+00 * singnx
  go to 240

230 continue

  fextrp = h2next
  h2next = 4.0d+00 * h2next

240 continue
 
  do i = it, lm1
    if ( abs ( dif(i+1) ) > tabtlm ) then
      r(i+1) = dif(i) / dif(i+1)
    else
      r(i+1) = 0.0d+00
    end if
  end do
 
  h2tfex = -h2tol*fextrp
  if ( r(l) - fextrp < h2tfex ) go to 270
  if ( r(l-1) - fextrp < h2tfex ) go to 270
  errer = astep * abs ( dif(l) )
  fextm1 = fextrp - 1.0d+00
  do i = it, l
    ait(i) = ait(i)+dif(i) / fextm1
    dif(i) = ait(i)-ait(i-1)
  end do
 
  go to 190
 
270 continue

  fextrp = dmax1(prever/errer,aitlow)
  prever = errer
  if (l<5) go to 40
  if (l-it>2.and.istage<mxstge) go to 370
  if (errer/fextrp**(mxtbl-l)<ergoal) go to 40
  go to 370
 
280 continue

  if ( errer > ergoal ) go to 370
  diff = abs ( t(1,l) ) * 2.0d+00 * fn
  go to 340
 
290 continue

  slope = (fend-fbeg) * 2.0d+00
  fbeg2 = fbeg * 2.0d+00
 
  do i = 1, 4
    diff = abs ( func ( beg + rn(i) * step ) - fbeg2 - rn(i) * slope )
    if ( diff > tabtlm ) go to 330
  end do
 
  go to 340
 
310 continue

  slope = (fend-fbeg)*2.0d+00
  fbeg2 = fbeg*2.0d+00
  i = 1
 
320 continue

  diff = abs ( func(beg+rn(i)*step) - fbeg2 - rn(i) * slope )
 
330 continue

  errer = dmax1( errer, astep * diff )
  if (errer > ergoal) go to 380
  i = i+1
  if ( i <= 4 ) go to 320
  ind = 3
 
340 continue

  result = result+vint
  error = error+errer
 
350 continue

  if (right) go to 360
  istage = istage-1
  if (istage==0) return
  reglar = reglsv(istage)
  beg = begin(istage)
  end = finis(istage)
  curest = curest-est(istage+1)+vint
  iend = ibeg-1
  fend = ts(iend)
  ibeg = ibegs(istage)
  go to 400
 
360 continue

  curest = curest+vint
  stage = stage*2.0d+00
  iend = ibeg
  ibeg = ibegs(istage)
  end = beg
  beg = begin(istage)
  fend = fbeg
  fbeg = ts(ibeg)
  go to 10
 
370 continue

  reglar = .true.
 
380 continue
 
  if ( istage == mxstge ) then
    ind = 5
    result = dble(curest+vint)
    return
  end if
 
390 continue

  if (right) go to 410
  reglsv(istage+1) = reglar
  begin(istage) = beg
  ibegs(istage) = ibeg
  stage = stage/2.0d+00

400 continue

  right = .true.
  beg = (beg+end)/2.0d+00
  ibeg = (ibeg+iend)/2
  ts(ibeg) = ts(ibeg) / 2.0d+00
  fbeg = ts(ibeg)
  go to 20

410 continue

  nnleft = ibeg-ibegs(istage)
  if (end+nnleft>=mxts) go to 440
  iii = ibegs(istage)
  ji = iend
  do i = iii, ibeg
    ji = ji+1
    ts(ji) = ts(i)
  end do
 
  do i = ibeg, ji
    ts(iii) = ts(i)
    iii = iii+1
  end do
 
  iend = iend+1
  ibeg = iend-nnleft
  fend = fbeg
  fbeg = ts(ibeg)
  finis(istage) = end
  end = beg
  beg = begin(istage)
  begin(istage) = end
  reglsv(istage) = reglar
  istage = istage+1
  reglar = reglsv(istage)
  est(istage) = vint
  curest = curest+est(istage)
  go to 10

440 continue

  ind = 4

460 continue

  result = dble(curest+vint)

  return
end subroutine rombe

FUNCTION hfunc(X)
 IMPLICIT NONE
 double precision  :: hfunc
 double precision, INTENT(IN) :: x
 hfunc=cos(x)/x
END FUNCTION hfunc

FUNCTION cosi(X)
 IMPLICIT NONE
 double precision  :: cosi
 double precision, INTENT(IN) :: x
 double precision  :: error
 integer :: i
 call rombe (hfunc,1.d+5,x,1.d-15,1.d-15,error,cosi,i)
 !print *,'cosi error: ',error
 !print *,'cosi index: ',i
!          ( func, a, b, aserr, relerr, error, result, ind )
END FUNCTION cosi

subroutine cosii(X,cosi)
 IMPLICIT NONE
 double precision  :: cosi
 double precision, INTENT(IN) :: x
 double precision  :: error
 integer :: i
 call rombe (hfunc,1.d+5,x,1.d-15,1.d-15,error,cosi,i)
 print *,'cosi val: ',cosi
 print *,'cosi error: ',error
 print *,'cosi index: ',i
      !    ( func, a, b, aserr, relerr, error, result, ind )
END subroutine cosii

FUNCTION func(X)
 IMPLICIT NONE
 double precision  :: func
 double precision, INTENT(IN) :: x
 func=x**2
END FUNCTION func

function xtoy(x,a,b)
 IMPLICIT NONE
 double precision, dimension(:)  :: x
 double precision, dimension(size(x,1))  :: xtoy
 double precision, INTENT(IN) :: a,b
 xtoy=2.d0*x-(a + b)
 xtoy=xtoy/(b-a)
end function xtoy

function ytoconfp(y)
 IMPLICIT NONE
 double precision, dimension(:)  :: y
 double precision, dimension(size(y,1))  :: ytoconfp
 ytoconfp=(1.d0/(1.d0-y**2))-1.d0
end function ytoconfp

function linearfit(y,bas)
 IMPLICIT NONE
 double precision, dimension(:)  :: y
 double precision, dimension(:,:)  :: bas
 double precision, dimension(size(bas,2))  :: linearfit
 double precision, dimension(size(bas,2),size(bas,2)) :: Mm
 double precision, dimension(size(bas,2)) :: ys
 double precision, allocatable,dimension(:)       :: WORK
 integer                   :: LDA,LWORK,INFO
 integer, allocatable      :: IPIV(:)

 integer :: n,m
 n=size(bas,1)
 m=size(bas,2)
 Mm=MATMUL(transpose(bas), bas)
 ys=MATMUL(transpose(bas),y)
 LDA=m
 LWORK=m
 allocate(WORK(LWORK))
 allocate(IPIV(m))
 call DGETRF(m, m, Mm, LDA, IPIV, INFO)
 call DGETRI(m,Mm,LDA,IPIV,WORK,LWORK,INFO)
 deallocate(WORK)
 deallocate(IPIV)
 linearfit=MATMUL(Mm,ys)
end function linearfit


FUNCTION fac(n) ! factorial
        INTEGER  fac
	INTEGER, INTENT(IN) :: n
       	INTEGER :: i
        INTEGER :: harray(n)
        if (n.eq.0) then
         fac=1 
        else    
	 harray = (/(i,i=1,n)/)
         fac=product(harray)
        end if
END FUNCTION fac

FUNCTION Cc(k,l)      ! help for IIs
      COMPLEX  :: Cc            ! function type  (kind = 1)
      INTEGER, INTENT( IN ) :: k,l
      Cc= 2*Ii**(k-1)*(1+(-1)**(k+l-1))
      Cc = CC/(k*pi)
END FUNCTION Cc

FUNCTION IIs(k,l)      
      ! (* \int_{-1}^{1} E^(I Pi/2*k (x+1)) *x^l dx=*)
      COMPLEX  :: IIs            ! function type  (kind = 1)
      INTEGER, INTENT( IN ) :: k,l
      INTEGER :: p
      COMPLEX  :: harray(l+1)
      if (k.eq.0) then 
       IIs=(1.d0+(-1)**l)/(1.d0+l)
      else
       harray=(/( ((2*Ii)/( k*pi ))**p*fac(l)/fac(l-p)*Cc(k, l - p), p=0,l) /)
       IIs=Sum(harray)
      end if 
END FUNCTION IIs

FUNCTION Jis(k,j,l) 
      ! Jis=\int_ {-1}^{1} Sin[Pi/2*k (x + 1)]*x^l Sin[ j Pi/2 (x + 1)] dx 
      DOUBLE PRECISION  :: Jis            ! function type  (kind = 1)
      COMPLEX  :: cJis
      INTEGER, INTENT( IN ) :: k,l,j
      cJis=Iis(k-j, l)*((-1)**j+(-1)**(k+l))-Iis(k+j,l)*(1 + (-1)**(j+k+l))
      Jis=DBLE(Ii**(j + k)/4*cJis)
END FUNCTION Jis

FUNCTION Kis(k,j,i) 
      ! Kis=\int_ {-1}^{1} Sin[Pi/2*i (x + 1)]*Sin[Pi/2*j (x + 1)] Sin[ k Pi/2 (x + 1)] dx 
      DOUBLE PRECISION  :: Kis            ! function type  (kind = 1)
      INTEGER, INTENT( IN ) :: i,j,k
      INTEGER               :: h
      h=(i-j-k)*(i-j+k)*(i+j-k)*(i+j+k)
      if (h==0) then
       Kis=0.0D0
      else
       Kis=4.0D0*i*j*k/pi/h
       h=(-1)**(i+j+k)-1
       Kis=DBLE(Kis*h)
      end if
END FUNCTION Kis

FUNCTION Cfp(i,j) 
      ! Cfp=\int_ {-1}^{1} Sin[(x+1)*pi/2*i]*Sin[(x+1)*pi/2*j]/(1 - x^2) dx
      DOUBLE PRECISION  :: Cfp,hef,ci,si            ! function type  (kind = 1)    
      INTEGER, INTENT(IN ) :: i,j   
      Cfp=0.d0
      if (MOD(iabs(i-j),2)==0) then
       Cfp=1.d0/2.d0
       call cisi(iabs(j+i)*dpi,ci,si)
       hef=-ci
       if (i-j==0) then !*EulerGamma + Log[2 Pi*i]
        hef=hef+egam+dlog(2*dpi*i)
       else !*(Log[(i+j)/Abs[(j-i)]]+CosIntegral[Abs[j-i]*Pi]]        
        call cisi(iabs(j-i)*dpi,ci,si)
        hef=hef+dlog((i+j)/dble(iabs(i-j)))+ci
       end if
      end if
      Cfp=Cfp*hef
END FUNCTION Cfp

! part splines

function fum(inx,hi) ! subfunction of naturalspine 
                     ! assigns rows of matrix 
 IMPLICIT NONE
 integer :: inx,n
 double precision, dimension(:)       :: hi
 double precision, dimension(size(hi,1)-1 )       :: fum
 n=size(hi,1)
 fum=0
 fum(inx)=2*(hi(inx)+hi(inx+1))
 if (inx>1) fum(inx-1)=hi(inx)
 if (inx<n-1) fum(inx+1)=hi(inx)
end function fum

function naturalsplin(datal)
 IMPLICIT NONE
 double precision, dimension(:,:)                     :: datal
 double precision, dimension(size(datal,1)-1,6)       :: naturalsplin
 double precision, dimension(size(datal,1)-1)         :: hi, dyi, ddyi
 double precision, dimension(size(datal,1))           :: ldyi
 double precision, dimension(size(datal,1)-2,size(datal,1)-2) :: ma
 double precision, allocatable,dimension(:)           :: WORK
 integer                   :: LDA,LWORK,INFO
 integer, allocatable      :: IPIV(:)
 integer :: npk,n,m,i,j
 npk=SIZE(datal,1) 
 m=npk-2
 hi=datal(2:npk,1)-datal(1:npk-1,1)
 dyi=datal(2:npk,2)-datal(1:npk-1,2)
 ddyi=dyi/hi
 ddyi=ddyi(2:npk-1)-ddyi(1:npk-2)
 ma=0;
 do i=1,npk-2
   ma(i,:)=fum(i,hi)
 enddo 
 LDA=m ;  LWORK=m
 allocate(WORK(LWORK))
 allocate(IPIV(m))
 call DGETRF(m, m, ma, LDA, IPIV, INFO)
 call DGETRI(m,ma,LDA,IPIV,WORK,LWORK,INFO)
 deallocate(WORK)
 deallocate(IPIV)
 ldyi=0.d0;
 ldyi(2:npk-1)=MATMUL(ma,6.d0*ddyi)
 do i=1,npk-1
  naturalsplin(i,1)=(ldyi(i+1 )-ldyi(i ))/(6.d0*hi(i))
  naturalsplin(i,3)=(datal(i+1,2)-datal(i,2))/hi(i)-(ldyi(i+1)+2.d0*ldyi(i))*hi(i)/6.d0
  naturalsplin(i,2)=ldyi(i)/2.d0
  naturalsplin(i,4)=datal(i,2) 
  naturalsplin(i,5)=datal(i,1)
  naturalsplin(i,6)=datal(i+1,1)
 enddo
end function naturalsplin


function splin(x,cfs) ! evaluate interpolation stored in cfs at x
 IMPLICIT NONE
 integer :: npk,i
 double precision, dimension(:,:)       :: cfs
 double precision       :: splin,x
 splin=0.d0
 npk=size(cfs,1)
 do i=1,npk
  if (x.le.cfs(i,6).and.x.ge.cfs(i,5)) then  !
   splin=x-cfs(i,5)
   splin=splin**3*cfs(i,1)+splin**2*cfs(i,2)+splin**1*cfs(i,3)+cfs(i,4)
   return
  end if
 enddo
end function splin


recursive function kr(n,wdk,a,xi,xii) result(krr)
 IMPLICIT NONE
 complex :: krr
 integer :: n
 double precision       :: wdk,a,xi,xii
 complex,parameter :: ci=(0.0,1.0)
 if (n.gt.0) then  !
  krr=((xii**n*exp(ci*wdk*(xii-a))-xi**n*exp(ci*wdk*(xi-a)))-n*kr(n-1,wdk,a,xi,xii))/(ci*wdk)
 else  
  krr=(exp(ci*wdk*(xii-a))- exp(ci*wdk*(xi-a)))/(ci*wdk)
 end if 
end function kr



function ka(n,wdk,a,xi,xii) ! 
 IMPLICIT NONE
 complex :: ka
 integer :: n
 double precision       :: wdk,a,xi,xii
 if (.not.dabs(wdk)==0.d0) then  !
  ka=kr(n,wdk,a,xi,xii)
 else 
  ka=(xii**(n+1)-xi**(n+1))/(n+1)
 end if 
end function ka

function li(n,k,ks,a,w,xi,xii) ! 
 IMPLICIT NONE
 double precision ::  a,w,xi,xii,li
 integer :: n,k,ks,tk,dk
 dk=k-ks ; tk=k+ks
 li=dble(ka(n,w*dk,a,xi,xii)+ka(n,(-1.d0)*w*dk,a,xi,xii)-ka(n,w*tk,a,xi,xii)-ka(n,(-1.d0)*w*tk,a,xi,xii))/4.d0
end function li

function ish(cfp,xi,xii,k,ks,a,b) ! 
 IMPLICIT NONE
 double precision,dimension(4) :: hlis,hliss
 integer              :: k,ks,i
 double precision     :: ish, a,b,xi,xii
 double precision,dimension(:) :: cfp
  hliss(1)=DOT_PRODUCT((/( (-xi)**(4-i),i=1,4) /),cfp(1:4))  
  hliss(2)=DOT_PRODUCT((/3.d0*xi**2,-2.d0*xi,1.d0/),cfp(1:3)) 
  hliss(3)=DOT_PRODUCT((/-3.d0*xi,1.d0/) ,cfp(1:2)) 
  hliss(4)=cfp(1)
  hlis=(/( li(i,k,ks,a,pi/(b-a),xi,xii),i=0,3)/)
  ish=DOT_PRODUCT(hlis,hliss)/(b-a)  
end function ish


function is(cfs,k,ks,a,b) ! 
 IMPLICIT NONE
 double precision   :: is
 integer            :: n,i,k,ks
 double precision   :: a,b
 double precision,dimension(:,:) :: cfs 
 n=size(cfs,1)
 is=DBLE(  sum( (/( ish(cfs(i,1:4),cfs(i,5),cfs(i,6),k,ks,a,b)  ,i=1,n )/)) )
end function is




! end part splines

! part analysis


function EigenVecsToFun(x,a,b,energies,cfs,temp,mmode) ! EigenVecsToFun
 ! for one mode calculates the probability at amplitude x
 ! mode = 0: suitable for -1 < x < 1
 ! mode = 1: suitable for  a < x < b
 IMPLICIT NONE
 double precision   :: is,EigenVecsToFun,temp,x,kap=dpi/2.d0
 integer            :: i,k,j,nocc,nenergies,ncfs,mode
 integer, optional  :: mmode 
 double precision   :: a,b,nnorm=1.d0
 double precision,dimension(:) :: energies
 double precision,dimension(:,:) :: cfs 
 double precision,dimension(:),allocatable :: occ,basf,lenergies 
    if(present(mmode))then
        mode=mmode
    else
        mode=0
    endif
  if (mode.eq.1) then 
   kap = pi/(b - a) ; nnorm=sqrt(2/(b-a))
  end if
  nenergies=size(energies); ncfs=size(cfs(1,:))
  allocate(occ(nenergies),basf(ncfs),lenergies(nenergies))
  occ=0.d0; lenergies=energies-energies(1)
  occ = exp((-1.d0)*lenergies(1:nenergies)*harteV/2.d0/(kbolzev*temp) ); 
  occ = occ/sum(occ)
  nocc = count(occ.gt.10.d-10)
  basf=nnorm*(/ ( sin(kap*(x + 1)*i) ,i=1,ncfs)/) !
  eigenvecstofun=sum( (/ ( DOT_PRODUCT(cfs(j,1:ncfs),basf)**2*occ(j),j=1,nocc)  /) )
  deallocate(occ,basf,lenergies)
end function EigenVecsToFun




function FreeE(emode,temparr,ntemp) ! 
 IMPLICIT NONE
 double precision   :: x
 double precision,dimension(ntemp) :: FreeE
 integer            :: i,k,j, nstat,nmode,ntemp,nocc
 double precision,dimension(:,:) :: emode
 double precision,dimension(:) :: temparr 
 double precision,dimension(:,:),allocatable ::  lenergies
 double precision,dimension(:),allocatable :: occ

  nstat=size(emode(1,:))
  nmode=size(emode(:,1))
  allocate(occ(nstat),lenergies(nmode,nstat))
  FreeE=0.d0; occ=0.d0; i=1 ; nocc=1;
  
  do while (i.le.nmode) 
   lenergies(i,1:nstat)=emode(i,1:nstat)-emode(i,1)
!   j=1
!   do while (j.le.nmode) 
    k=1
    do while (k.le.ntemp) 
     occ(1:nstat)=exp((-1.d0)*lenergies(i,1:nstat)*harteV/2.d0/(kbolzev*temparr(k)) ); 
     occ(1:nstat) = occ(1:nstat)/sum(occ(1:nstat))
     nocc=count(occ(1:nstat).gt.10.d-10)!maxval((/ nocc,count(occ(k,i,1:nstat).gt.10.d-10) /) )
     FreeE(k)=FreeE(k)-kbolzev*temparr(k)* log( sum( exp((-1.d0)*lenergies(i,1:nocc)*harteV/2.d0/(kbolzev*temparr(k))) ))
 !DOT_PRODUCT( occ( 1:nocc) ,lenergies(i,1:nocc)) !nstat+nmode  ! 
     k=k+1
!    end do
!   j=j+1
  end do
  i=i+1.

  end do
  deallocate(occ,lenergies )
end function FreeE

!FreeE(emode(1:nmode,1:nstat),temparr(1:ntemp),ntemp)
!VibEllips(emode(1:nmode,1:nstat),mmode(i,j),sigmode(i,j),pmode(1:nmode,1:nat,1:3),temparr(1:ntemp),ntemp)

function VibEllips(emode,mmode,sigmode,pmode,temparr,ntemp,nat) ! 
 IMPLICIT NONE
 double precision   :: x
 double precision,dimension(ntemp,nat,9) :: VibEllips
 integer            :: i,k,j, nstat,nmode,ntemp,nocc,itemp,imode,iat ,nat
 double precision,dimension(3,3) :: ha
 double precision,dimension(:,:) :: emode,mmode,sigmode
 double precision,dimension(:,:,:) :: pmode
 double precision,dimension(:) :: temparr 
 double precision,dimension(:,:),allocatable ::  lenergies
 double precision,dimension(:),allocatable :: occ

  nstat=size(emode(1,:))
  nmode=size(emode(:,1)) 
  allocate(occ(nstat),lenergies(nmode,nstat))
  VibEllips=0.d0;  imode=1 ; 
 ! VibEllips(1:ntemp,1:nat,1:9)=iat+ntemp+nstat+nmode
 ! VibEllips(1:ntemp,1,1)=nat
 ! VibEllips(1:ntemp,1,2)=ntemp
 ! VibEllips(1:ntemp,1,3)=nstat
 ! VibEllips(1:ntemp,1,4)=nmode
  do while (imode.le.nmode)  !nmode
   lenergies(imode,1:nstat)=emode(imode,1:nstat)-emode(imode,1)
   itemp=1
   do while (itemp.le.ntemp) 
    occ(1:nstat)=exp((-1.d0)*lenergies(i,1:nstat)*harteV/2.d0/(kbolzev*temparr(itemp)) ); 
    occ(1:nstat) = occ(1:nstat)/sum(occ(1:nstat))
    nocc=count(occ(1:nstat).gt.1.d-9)!maxval((/ nocc,count(occ(k,i,1:nstat).gt.10.d-10) /) )
    iat=1 ; ha=0.d0
    do while (iat.le.nat)
     ha=spread(pmode(imode,iat,1:3),1,3) !*sqrt(dot_product(sigmode(imode,1:nocc),occ(1:nocc)))
     VibEllips(itemp,iat,1:9)=VibEllips(itemp,iat,1:9)+reshape(ha*transpose(ha), (/ 9 /) )
     iat=iat+1
    end do
    itemp=itemp+1
   end do
   imode=imode+1
  end do
  deallocate(occ,lenergies )
end function VibEllips

function MeV(xarr,aarr,barr,energiesarr,cfsarr,temp,mmode) ! 
 ! calculates the probability at amplitude x as product of all modes 
 ! mode = 0: suitable for -1 < x < 1
 ! mode = 1: suitable for  a < x < b
 IMPLICIT NONE
 double precision   :: MeV
 integer            :: i,k,j,nmods,info,mode
 integer, optional  :: mmode 
 double precision   :: temp
 integer,dimension(3) :: ipiv
 double precision,dimension(:,:,:) :: cfsarr 
 double precision,dimension(:,:) :: energiesarr
 double precision,dimension(:) :: xarr,aarr,barr 
    if(present(mmode))then
        mode=mmode
    else
        mode=0
    endif
 nmods=size(xarr)
 MeV=product( (/ ( eigenvecstofun(xarr(j),aarr(j),barr(j),energiesarr(j,:),cfsarr(j,:,:),temp,mode),j=1,nmods)  /))
end function MeV

function PixToDen(pix,aarr,barr,energiesarr,cfsarr,temparr,ntemp,modsn,pos,dpix) ! 
 ! calculates the probability at amplitude x as product of all modes 
 IMPLICIT NONE
 integer            :: in1,in2,in3,iat,i,k,j,nmods,ntemp,info,nat
 integer,dimension(3)::ipiv
 double precision   :: temp
 double precision,dimension(3) :: pix,ampll
 double precision,dimension(3,3) :: A
 double precision,dimension(:,:,:) :: cfsarr,modsn
 double precision,dimension(:,:) :: energiesarr,pos
 double precision,dimension(ntemp) :: PixToDen
 double precision,dimension(:) :: aarr,barr ,temparr, dpix
 double precision,dimension(:),allocatable :: lampe
 nmods=size(aarr)
 nat=size(pos(:,1))
 !PixToDen(1)=nmods*1.d0+nat 
 PixToDen=0.d0
 allocate(lampe(nmods))
  in1=1
  do while (in1.lt.nmods-1)
   in2 = in1+1
   do while (in2.lt.nmods)
    in3 = in2+1
    do while (in3.le.nmods)
     iat = 1
     do while (iat.le.nat)
      ampll=pix-pos(iat,1:3)
      !PixToDen(1) = PixToDen(1) + norm2(modsn(in1,iat,1:3))+norm2(modsn(in2,iat,1:3))+norm2(modsn(in3, iat,1:3))
      if (norm2(modsn(in1,iat,1:3))+norm2(modsn(in2,iat,1:3))+norm2(modsn(in3, iat,1:3)).gt.norm2(ampll)-norm2(dpix)) then  
        A=transpose(reshape(   (/ modsn(in1, iat,1:3), modsn(in2, iat,1:3), modsn(in3, iat,1:3) /),(/ 3,3 /)  ))
        ipiv=0 ; info=0
        call dgesv (3, 3, A , 3, ipiv, ampll, 3, info)
       !     PixToDen(1) = PixToDen(1) + info !abs(ampll(1))+abs(ampll(2))+abs(ampll(3))
        if (abs(ampll(1)).lt.1.d0.and.abs(ampll(2)).lt.1.d0.and.abs(ampll(3)).lt.1.d0.and.info.eq.0) then
        lampe = 0.d0
        lampe(in1) = ampll(1); lampe(in2) = ampll(2); lampe(in3) = ampll(3);
        k=1
        do while (k.le.ntemp)
        PixToDen(k) = PixToDen(k) + MeV(lampe,aarr,barr,energiesarr,cfsarr, temparr(k),0) 
     !  PixToDen(k) = PixToDen(k) + abs(ampll(1)) !+ abs(ampll(2))+ abs(ampll(3))
      !  PixToDen(k) = PixToDen(k) + abs(ampll(1)).lt.1.d0.and.abs(ampll(2)).lt.1.d0.and.abs(ampll(3)).lt.1.d0 ! abs(ampll(1)) !+ abs(ampll(2))+ abs(ampll(3))
    !      return
         k=k+1
        end do
         end if
      end if
      iat = iat + 1
     end do
     in3 = in3 + 1
    end do
    in2 = in2 + 1
   end do
   in1 = in1 + 1
  end do
  deallocate(lampe)
end function PixToDen



! end part analysis

    SUBROUTINE cisi(x,ci,si)
     INTEGER MAXIT
     double precision ci,si,x,EPS,PIBY2,FPMIN,TMIN
     PARAMETER (EPS=5.d-16,MAXIT=1000,PIBY2=dpi/2.d0, FPMIN=1.0d-300,TMIN=2.)
   !     The cosine and sine integrals Ci(x) and Si(x). 
   !     Parameters: EPS is the relative error, or absolute error near a zero of Ci(x); 
   !     MAXIT is the maximum number of iterations allowed; PIBY2 = π/2; FPMIN is a number
   !     near the smallest representable floating-point number; TMIN is the dividing line between
   !     using the series and continued fraction.
     INTEGER i,k
     double precision a,err,fact,sign,sum,sumc,sums,t,term,absc
     COMPLEX*16 h,b,c,d,del
     LOGICAL odd
     absc(h)=abs(real(h))+abs(aimag(h)) ! Statement function.
     t=abs(x)
     if(t.eq.0.)then ! Special case.
      si=0.d0
      ci=-1.d0/FPMIN
      return
     endif
     if(t.gt.TMIN)then ! Evaluate continued fraction by modified Lentz’s method
      b=cmplx(1.d0,t) 
      c=1.d0/FPMIN
      d=1.d0/b
      h=d
      do i=2,MAXIT
       a=-(i-1.d0)**2
       b=b+2.d0
       d=1.d0/(a*d+b) 
       c=b+a/c
       del=c*d
       h=h*del
       if(absc(del-1.d0).lt.EPS) goto 1
      enddo 
      print *, 'cf failed in cisi'
1     continue
      h=cmplx(cos(t),-sin(t))*h
      ci=-real(h)
      si=PIBY2+aimag(h)
     else                      ! Evaluate both series simultaneously.
      if(t.lt.sqrt(FPMIN))then ! Special case: avoid failure of convergence test
       sumc=0.d0                 ! because of underflow.
       sums=t
      else
       sum=0.d0
       sums=0.d0
       sumc=0.d0
       sign=1.d0
       fact=1.d0
       odd=.true.
  do k=1,MAXIT
    fact=fact*t/k
    term=fact/k
    sum=sum+sign*term
    err=term/abs(sum)
    if(odd)then
     sign=-sign
     sums=sum
     sum=sumc
    else
     sumc=sum
     sum=sums
    endif
    if(err.lt.EPS)goto 2
    odd=.not.odd
  enddo 
  print *,'maxits exceeded in cisi'
 endif
2 si=sums
  ci=sumc+log(t)+egam
 endif
 if(x.lt.0.)si=-si
 return
END SUBROUTINE cisi



end module eigenstate_utils
